#include "angband.h"
#include "grid.h"
#include "generate.h"
#include "rooms.h"
#include "rooms-pitnest.h"
#include "monster-hook.h"



#define NUM_NEST_MON_TYPE 64 /*!<nest̎ʐ */

/*! pit/nest^typedef */
typedef struct vault_aux_type vault_aux_type;

/*! pit/nest^̍\̒` */
struct vault_aux_type
{
	cptr name;
	bool(*hook_func)(MONRACE_IDX r_idx);
	void(*prep_func)(void);
	int level;
	int chance;
};

/*! nestID` /  Nest types code */
#define NEST_TYPE_CLONE        0
#define NEST_TYPE_JELLY        1
#define NEST_TYPE_SYMBOL_GOOD  2
#define NEST_TYPE_SYMBOL_EVIL  3
#define NEST_TYPE_MIMIC        4
#define NEST_TYPE_LOVECRAFTIAN 5
#define NEST_TYPE_KENNEL       6
#define NEST_TYPE_ANIMAL       7
#define NEST_TYPE_CHAPEL       8
#define NEST_TYPE_UNDEAD       9

/*! pitID` / Pit types code */
#define PIT_TYPE_ORC           0
#define PIT_TYPE_TROLL         1
#define PIT_TYPE_GIANT         2
#define PIT_TYPE_LOVECRAFTIAN  3
#define PIT_TYPE_SYMBOL_GOOD   4
#define PIT_TYPE_SYMBOL_EVIL   5
#define PIT_TYPE_CHAPEL        6
#define PIT_TYPE_DRAGON        7
#define PIT_TYPE_DEMON         8
#define PIT_TYPE_DARK_ELF      9






/*! ʏpit̃X^[̍\ID / Race index for "monster pit (clone)" */
static int vault_aux_race;

/*! PV{pit̎wV{ / Race index for "monster pit (symbol clone)" */
static char vault_aux_char;

/*! uXɊÂhSpit}XN / Breath mask for "monster pit (dragon)" */
static u32b vault_aux_dragon_mask4;


/*!
* @brief X^[Vault̍ŒKv𖞂Ă邩Ԃ /
* Helper monster selection function
* @param r_idx mFX^[푰ID
* @return Vault̍ŒKv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_simple(MONRACE_IDX r_idx)
{
	/* Okay */
	return (vault_monster_okay(r_idx));
}


/*!
* @brief X^[[[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (jelly)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_jelly(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if ((r_ptr->flags2 & RF2_KILL_BODY) && !(r_ptr->flags1 & RF1_NEVER_BLOW)) return (FALSE);

	/* Also decline evil jellies (like death molds and shoggoths) */
	if (r_ptr->flags3 & (RF3_EVIL)) return (FALSE);

	/* Require icky thing, jelly, mold, or mushroom */
	if (!my_strchr("ijm,", r_ptr->d_char)) return (FALSE);

	/* Okay */
	return (TRUE);
}

/*!
* @brief X^[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (animal)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_animal(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require "animal" flag */
	if (!(r_ptr->flags3 & (RF3_ANIMAL))) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[Afbhnest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (undead)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_undead(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require Undead */
	if (!(r_ptr->flags3 & (RF3_UNDEAD))) return (FALSE);

	/* Okay */
	return (TRUE);
}

/*!
* @brief X^[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (chapel)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_chapel_g(MONRACE_IDX r_idx)
{
	static int chapel_list[] = {
		MON_NOV_PRIEST, MON_NOV_PALADIN, MON_NOV_PRIEST_G, MON_NOV_PALADIN_G,
		MON_PRIEST, MON_JADE_MONK, MON_IVORY_MONK, MON_ULTRA_PALADIN,
		MON_EBONY_MONK, MON_W_KNIGHT, MON_KNI_TEMPLAR, MON_PALADIN,
		MON_TOPAZ_MONK, 0 };

	int i;

	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if (r_ptr->flags3 & (RF3_EVIL)) return (FALSE);
	if ((r_idx == MON_A_GOLD) || (r_idx == MON_A_SILVER)) return (FALSE);

	/* Require "priest" or Angel */

	if (r_ptr->d_char == 'A') return TRUE;

	for (i = 0; chapel_list[i]; i++)
		if (r_idx == chapel_list[i]) return TRUE;

	return FALSE;
}

/*!
* @brief X^[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (kennel)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_kennel(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require a Zephyr Hound or a dog */
	if (!my_strchr("CZ", r_ptr->d_char)) return (FALSE);

	/* Okay */
	return (TRUE);
}

/*!
* @brief X^[~~bNnest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (mimic)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_mimic(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require mimic */
	if (!my_strchr("!$&(/=?[\\|", r_ptr->d_char)) return (FALSE);

	/* Okay */
	return (TRUE);
}

/*!
* @brief X^[PN[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (clone)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_clone(MONRACE_IDX r_idx)
{
	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	return (r_idx == vault_aux_race);
}


/*!
* @brief X^[׈V{N[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (symbol clone)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_symbol_e(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if ((r_ptr->flags2 & RF2_KILL_BODY) && !(r_ptr->flags1 & RF1_NEVER_BLOW)) return (FALSE);

	if (r_ptr->flags3 & (RF3_GOOD)) return (FALSE);

	/* Decline incorrect symbol */
	if (r_ptr->d_char != vault_aux_char) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[PǑV{N[nest̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster nest (symbol clone)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_symbol_g(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if ((r_ptr->flags2 & RF2_KILL_BODY) && !(r_ptr->flags1 & RF1_NEVER_BLOW)) return (FALSE);

	if (r_ptr->flags3 & (RF3_EVIL)) return (FALSE);

	/* Decline incorrect symbol */
	if (r_ptr->d_char != vault_aux_char) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[I[Npit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (orc)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_orc(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require orc */
	if (!(r_ptr->flags3 & RF3_ORC)) return (FALSE);

	/* Decline undead */
	if (r_ptr->flags3 & RF3_UNDEAD) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[gpit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (troll)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_troll(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require troll */
	if (!(r_ptr->flags3 & RF3_TROLL)) return (FALSE);

	/* Decline undead */
	if (r_ptr->flags3 & RF3_UNDEAD) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[lpit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (giant)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_giant(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require giant */
	if (!(r_ptr->flags3 & RF3_GIANT)) return (FALSE);

	if (r_ptr->flags3 & RF3_GOOD) return (FALSE);

	/* Decline undead */
	if (r_ptr->flags3 & RF3_UNDEAD) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[hSpit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (dragon)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_dragon(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* Require dragon */
	if (!(r_ptr->flags3 & RF3_DRAGON)) return (FALSE);

	/* Hack -- Require correct "breath attack" */
	if (r_ptr->flags4 != vault_aux_dragon_mask4) return (FALSE);

	/* Decline undead */
	if (r_ptr->flags3 & RF3_UNDEAD) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[pit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (demon)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_demon(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if ((r_ptr->flags2 & RF2_KILL_BODY) && !(r_ptr->flags1 & RF1_NEVER_BLOW)) return (FALSE);

	/* Require demon */
	if (!(r_ptr->flags3 & RF3_DEMON)) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief X^[Cpit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (lovecraftian)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_cthulhu(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	if ((r_ptr->flags2 & RF2_KILL_BODY) && !(r_ptr->flags1 & RF1_NEVER_BLOW)) return (FALSE);

	/* Require eldritch horror */
	if (!(r_ptr->flags2 & (RF2_ELDRITCH_HORROR))) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief pit/nest̊ƂȂP탂X^[߂ /
* @return Ȃ
*/
static void vault_prep_clone(void)
{
	/* Apply the monster restriction */
	get_mon_num_prep(vault_aux_simple, NULL);

	/* Pick a race to clone */
	vault_aux_race = get_mon_num(dun_level + 10);

	/* Remove the monster restriction */
	get_mon_num_prep(NULL, NULL);
}


/*!
* @brief pit/nest̊ƂȂ郂X^[V{߂ /
* @return Ȃ
*/
static void vault_prep_symbol(void)
{
	MONRACE_IDX r_idx;

	/* Apply the monster restriction */
	get_mon_num_prep(vault_aux_simple, NULL);

	/* Pick a race to clone */
	r_idx = get_mon_num(dun_level + 10);

	/* Remove the monster restriction */
	get_mon_num_prep(NULL, NULL);

	/* Extract the symbol */
	vault_aux_char = r_info[r_idx].d_char;
}

/*!
* @brief pit/nest̊ƂȂhS̎ނ߂ /
* @return Ȃ
*/
static void vault_prep_dragon(void)
{
	/* Pick dragon type */
	switch (randint0(6))
	{
		/* Black */
	case 0:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = RF4_BR_ACID;

		/* Done */
		break;
	}

	/* Blue */
	case 1:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = RF4_BR_ELEC;

		/* Done */
		break;
	}

	/* Red */
	case 2:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = RF4_BR_FIRE;

		/* Done */
		break;
	}

	/* White */
	case 3:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = RF4_BR_COLD;

		/* Done */
		break;
	}

	/* Green */
	case 4:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = RF4_BR_POIS;

		/* Done */
		break;
	}

	/* Multi-hued */
	default:
	{
		/* Restrict dragon breath type */
		vault_aux_dragon_mask4 = (RF4_BR_ACID | RF4_BR_ELEC |
			RF4_BR_FIRE | RF4_BR_COLD |
			RF4_BR_POIS);

		/* Done */
		break;
	}
	}
}


/*!
* @brief X^[_[NGtpit̐Kv𖞂Ă邩Ԃ /
* Helper function for "monster pit (dark elf)"
* @param r_idx mFX^[푰ID
* @return Kv𖞂ĂȂTRUEԂB
*/
static bool vault_aux_dark_elf(MONRACE_IDX r_idx)
{
	int i;
	static int dark_elf_list[] =
	{
		MON_D_ELF, MON_D_ELF_MAGE, MON_D_ELF_WARRIOR, MON_D_ELF_PRIEST,
		MON_D_ELF_LORD, MON_D_ELF_WARLOCK, MON_D_ELF_DRUID, MON_NIGHTBLADE,
		MON_D_ELF_SORC, MON_D_ELF_SHADE, 0,
	};

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return FALSE;

	/* Require dark elves */
	for (i = 0; dark_elf_list[i]; i++)
		if (r_idx == dark_elf_list[i]) return TRUE;

	/* Assume not */
	return FALSE;
}


/*!
* @brief _WɎw肳ꂽsbgzɃ_pit/nest^Cv߂
* @param l_ptr Iꂽpit/nestԂQƃ|C^
* @param allow_flag_mask pit/nest̃rbgz
* @return Iꂽpit/nestIDAIsꍇ-1ԂB
*/
static int pick_vault_type(vault_aux_type *l_ptr, s16b allow_flag_mask)
{
	int tmp, total, count;

	vault_aux_type *n_ptr;

	/* Calculate the total possibilities */
	for (n_ptr = l_ptr, total = 0, count = 0; TRUE; n_ptr++, count++)
	{
		/* Note end */
		if (!n_ptr->name) break;

		/* Ignore excessive depth */
		if (n_ptr->level > dun_level) continue;

		/* Not matched with pit/nest flag */
		if (!(allow_flag_mask & (1L << count))) continue;

		/* Count this possibility */
		total += n_ptr->chance * MAX_DEPTH / (MIN(dun_level, MAX_DEPTH - 1) - n_ptr->level + 5);
	}

	/* Pick a random type */
	tmp = randint0(total);

	/* Find this type */
	for (n_ptr = l_ptr, total = 0, count = 0; TRUE; n_ptr++, count++)
	{
		/* Note end */
		if (!n_ptr->name) break;

		/* Ignore excessive depth */
		if (n_ptr->level > dun_level) continue;

		/* Not matched with pit/nest flag */
		if (!(allow_flag_mask & (1L << count))) continue;

		/* Count this possibility */
		total += n_ptr->chance * MAX_DEPTH / (MIN(dun_level, MAX_DEPTH - 1) - n_ptr->level + 5);

		/* Found the type */
		if (tmp < total) break;
	}

	return n_ptr->name ? count : -1;
}

/*!
* @brief fobOɐꂽpit/nesť^o͂鏈
* @param type pit/nesť^ID
* @param nest TRUEȂnestAFALSEȂpit
* @return fobO\̎Qƃ|C^
* @details
* Hack -- Get the string describing subtype of pit/nest
* Determined in prepare function (some pit/nest only)
*/
static cptr pit_subtype_string(int type, bool nest)
{
	static char inner_buf[256] = "";

	inner_buf[0] = '\0'; /* Init string */

	if (nest) /* Nests */
	{
		switch (type)
		{
		case NEST_TYPE_CLONE:
			sprintf(inner_buf, "(%s)", r_name + r_info[vault_aux_race].name);
			break;
		case NEST_TYPE_SYMBOL_GOOD:
		case NEST_TYPE_SYMBOL_EVIL:
			sprintf(inner_buf, "(%c)", vault_aux_char);
			break;
		}
	}
	else /* Pits */
	{
		switch (type)
		{
		case PIT_TYPE_SYMBOL_GOOD:
		case PIT_TYPE_SYMBOL_EVIL:
			sprintf(inner_buf, "(%c)", vault_aux_char);
			break;
		case PIT_TYPE_DRAGON:
			switch (vault_aux_dragon_mask4)
			{
#ifdef JP
			case RF4_BR_ACID: strcpy(inner_buf, "(_)");   break;
			case RF4_BR_ELEC: strcpy(inner_buf, "()"); break;
			case RF4_BR_FIRE: strcpy(inner_buf, "(Ή)"); break;
			case RF4_BR_COLD: strcpy(inner_buf, "(C)"); break;
			case RF4_BR_POIS: strcpy(inner_buf, "()");   break;
			case (RF4_BR_ACID | RF4_BR_ELEC | RF4_BR_FIRE | RF4_BR_COLD | RF4_BR_POIS) :
				strcpy(inner_buf, "(F)"); break;
			default: strcpy(inner_buf, "(`)"); break;
#else
			case RF4_BR_ACID: strcpy(inner_buf, "(acid)");      break;
			case RF4_BR_ELEC: strcpy(inner_buf, "(lightning)"); break;
			case RF4_BR_FIRE: strcpy(inner_buf, "(fire)");      break;
			case RF4_BR_COLD: strcpy(inner_buf, "(frost)");     break;
			case RF4_BR_POIS: strcpy(inner_buf, "(poison)");    break;
			case (RF4_BR_ACID | RF4_BR_ELEC | RF4_BR_FIRE | RF4_BR_COLD | RF4_BR_POIS) :
				strcpy(inner_buf, "(multi-hued)"); break;
			default: strcpy(inner_buf, "(undefined)"); break;
#endif
			}
			break;
		}
	}

	return inner_buf;
}





/*
*! @brief nest̃X^[Xg\[g邽߂̊֐ /
*  Comp function for sorting nest monster information
*  @param u \[gΏ۔z|C^
*  @param v gp
*  @param a rΏێQID1
*  @param b rΏێQID2
*/
static bool ang_sort_comp_nest_mon_info(vptr u, vptr v, int a, int b)
{
	nest_mon_info_type *nest_mon_info = (nest_mon_info_type *)u;
	int w1 = nest_mon_info[a].r_idx;
	int w2 = nest_mon_info[b].r_idx;
	monster_race *r1_ptr = &r_info[w1];
	monster_race *r2_ptr = &r_info[w2];
	int z1, z2;

	/* Unused */
	(void)v;

	/* Extract used info */
	z1 = nest_mon_info[a].used;
	z2 = nest_mon_info[b].used;

	/* Compare used status */
	if (z1 < z2) return FALSE;
	if (z1 > z2) return TRUE;

	/* Compare levels */
	if (r1_ptr->level < r2_ptr->level) return TRUE;
	if (r1_ptr->level > r2_ptr->level) return FALSE;

	/* Compare experience */
	if (r1_ptr->mexp < r2_ptr->mexp) return TRUE;
	if (r1_ptr->mexp > r2_ptr->mexp) return FALSE;

	/* Compare indexes */
	return w1 <= w2;
}

/*!
* @brief nest̃X^[XgXbv邽߂̊֐ /
* Swap function for sorting nest monster information
* @param u XbvΏ۔z|C^
* @param v gp
* @param a XbvΏێQID1
* @param b XbvΏێQID2
*/
static void ang_sort_swap_nest_mon_info(vptr u, vptr v, int a, int b)
{
	nest_mon_info_type *nest_mon_info = (nest_mon_info_type *)u;
	nest_mon_info_type holder;

	/* Unused */
	(void)v;

	/* Swap */
	holder = nest_mon_info[a];
	nest_mon_info[a] = nest_mon_info[b];
	nest_mon_info[b] = holder;
}



/*!neste[u*/
static vault_aux_type nest_types[] =
{
#ifdef JP
{ "N[",     vault_aux_clone,    vault_prep_clone,   5, 3 },
{ "[[",       vault_aux_jelly,    NULL,               5, 6 },
{ "V{(P)", vault_aux_symbol_g, vault_prep_symbol, 25, 2 },
{ "V{()", vault_aux_symbol_e, vault_prep_symbol, 25, 2 },
{ "~~bN",     vault_aux_mimic,    NULL,              30, 4 },
{ "C",         vault_aux_cthulhu,  NULL,              70, 2 },
{ "",       vault_aux_kennel,   NULL,              45, 4 },
{ "",       vault_aux_animal,   NULL,              35, 5 },
{ "",         vault_aux_chapel_g, NULL,              75, 4 },
{ "Afbh",   vault_aux_undead,   NULL,              75, 5 },
{ NULL,           NULL,               NULL,               0, 0 },
#else
{ "clone",        vault_aux_clone,    vault_prep_clone,   5, 3 },
{ "jelly",        vault_aux_jelly,    NULL,               5, 6 },
{ "symbol good",  vault_aux_symbol_g, vault_prep_symbol, 25, 2 },
{ "symbol evil",  vault_aux_symbol_e, vault_prep_symbol, 25, 2 },
{ "mimic",        vault_aux_mimic,    NULL,              30, 4 },
{ "lovecraftian", vault_aux_cthulhu,  NULL,              70, 2 },
{ "kennel",       vault_aux_kennel,   NULL,              45, 4 },
{ "animal",       vault_aux_animal,   NULL,              35, 5 },
{ "chapel",       vault_aux_chapel_g, NULL,              75, 4 },
{ "undead",       vault_aux_undead,   NULL,              75, 5 },
{ NULL,           NULL,               NULL,               0, 0 },
#endif
};

/*!pite[u*/
static vault_aux_type pit_types[] =
{
#ifdef JP
{ "I[N",       vault_aux_orc,      NULL,               5, 6 },
{ "g",       vault_aux_troll,    NULL,              20, 6 },
{ "WCAg", vault_aux_giant,    NULL,              50, 6 },
{ "C",         vault_aux_cthulhu,  NULL,              80, 2 },
{ "V{(P)", vault_aux_symbol_g, vault_prep_symbol, 70, 1 },
{ "V{()", vault_aux_symbol_e, vault_prep_symbol, 70, 1 },
{ "",         vault_aux_chapel_g, NULL,              65, 2 },
{ "hS",     vault_aux_dragon,   vault_prep_dragon, 70, 6 },
{ "f[",     vault_aux_demon,    NULL,              80, 6 },
{ "_[NGt", vault_aux_dark_elf, NULL,              45, 4 },
{ NULL,           NULL,               NULL,               0, 0 },
#else
{ "orc",          vault_aux_orc,      NULL,               5, 6 },
{ "troll",        vault_aux_troll,    NULL,              20, 6 },
{ "giant",        vault_aux_giant,    NULL,              50, 6 },
{ "lovecraftian", vault_aux_cthulhu,  NULL,              80, 2 },
{ "symbol good",  vault_aux_symbol_g, vault_prep_symbol, 70, 1 },
{ "symbol evil",  vault_aux_symbol_e, vault_prep_symbol, 70, 1 },
{ "chapel",       vault_aux_chapel_g, NULL,              65, 2 },
{ "dragon",       vault_aux_dragon,   vault_prep_dragon, 70, 6 },
{ "demon",        vault_aux_demon,    NULL,              80, 6 },
{ "dark elf",     vault_aux_dark_elf, NULL,              45, 4 },
{ NULL,           NULL,               NULL,               0, 0 },
#endif
};




/*!
* @brief ^Cv5̕cnest𐶐 / Type 5 -- Monster nests
* @return Ȃ
* @details
* A monster nest is a "big" room, with an "inner" room, containing\n
* a "collection" of monsters of a given type strewn about the room.\n
*\n
* The monsters are chosen from a set of 64 randomly selected monster\n
* races, to allow the nest creation to fail instead of having "holes".\n
*\n
* Note the use of the "get_mon_num_prep()" function, and the special\n
* "get_mon_num_hook()" restriction function, to prepare the "monster\n
* allocation table" in such a way as to optimize the selection of\n
* "appropriate" non-unique monsters for the nest.\n
*\n
* Note that the "get_mon_num()" function may (rarely) fail, in which\n
* case the nest will be empty.\n
*\n
* Note that "monster nests" will never contain "unique" monsters.\n
*/
bool build_type5(void)
{
	POSITION y, x, y1, x1, y2, x2, xval, yval;
	int i;
	nest_mon_info_type nest_mon_info[NUM_NEST_MON_TYPE];

	monster_type align;

	cave_type *c_ptr;

	int cur_nest_type = pick_vault_type(nest_types, d_info[dungeon_type].nest);
	vault_aux_type *n_ptr;

	/* No type available */
	if (cur_nest_type < 0) return FALSE;

	n_ptr = &nest_types[cur_nest_type];

	/* Process a preparation function if necessary */
	if (n_ptr->prep_func) (*(n_ptr->prep_func))();

	/* Prepare allocation table */
	get_mon_num_prep(n_ptr->hook_func, NULL);

	align.sub_align = SUB_ALIGN_NEUTRAL;

	/* Pick some monster types */
	for (i = 0; i < NUM_NEST_MON_TYPE; i++)
	{
		MONRACE_IDX r_idx = 0;
		int attempts = 100;
		monster_race *r_ptr = NULL;

		while (attempts--)
		{
			/* Get a (hard) monster type */
			r_idx = get_mon_num(dun_level + 11);
			r_ptr = &r_info[r_idx];

			/* Decline incorrect alignment */
			if (monster_has_hostile_align(&align, 0, 0, r_ptr)) continue;

			/* Accept this monster */
			break;
		}

		/* Notice failure */
		if (!r_idx || !attempts) return FALSE;

		/* Note the alignment */
		if (r_ptr->flags3 & RF3_EVIL) align.sub_align |= SUB_ALIGN_EVIL;
		if (r_ptr->flags3 & RF3_GOOD) align.sub_align |= SUB_ALIGN_GOOD;

		nest_mon_info[i].r_idx = (s16b)r_idx;
		nest_mon_info[i].used = FALSE;
	}

	/* Find and reserve some space in the dungeon.  Get center of room. */
	if (!find_space(&yval, &xval, 11, 25)) return FALSE;

	/* Large room */
	y1 = yval - 4;
	y2 = yval + 4;
	x1 = xval - 11;
	x2 = xval + 11;

	/* Place the floor area */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		for (x = x1 - 1; x <= x2 + 1; x++)
		{
			c_ptr = &cave[y][x];
			place_floor_grid(c_ptr);
			c_ptr->info |= (CAVE_ROOM);
		}
	}

	/* Place the outer walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		c_ptr = &cave[y][x1 - 1];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y][x2 + 1];
		place_outer_grid(c_ptr);
	}
	for (x = x1 - 1; x <= x2 + 1; x++)
	{
		c_ptr = &cave[y1 - 1][x];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y2 + 1][x];
		place_outer_grid(c_ptr);
	}


	/* Advance to the center room */
	y1 = y1 + 2;
	y2 = y2 - 2;
	x1 = x1 + 2;
	x2 = x2 - 2;

	/* The inner walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		c_ptr = &cave[y][x1 - 1];
		place_inner_grid(c_ptr);
		c_ptr = &cave[y][x2 + 1];
		place_inner_grid(c_ptr);
	}

	for (x = x1 - 1; x <= x2 + 1; x++)
	{
		c_ptr = &cave[y1 - 1][x];
		place_inner_grid(c_ptr);
		c_ptr = &cave[y2 + 1][x];
		place_inner_grid(c_ptr);
	}
	for (y = y1; y <= y2; y++)
	{
		for (x = x1; x <= x2; x++)
		{
			add_cave_info(y, x, CAVE_ICKY);
		}
	}

	/* Place a secret door */
	switch (randint1(4))
	{
	case 1: place_secret_door(y1 - 1, xval, DOOR_DEFAULT); break;
	case 2: place_secret_door(y2 + 1, xval, DOOR_DEFAULT); break;
	case 3: place_secret_door(yval, x1 - 1, DOOR_DEFAULT); break;
	case 4: place_secret_door(yval, x2 + 1, DOOR_DEFAULT); break;
	}

	msg_format_wizard(CHEAT_DUNGEON, _("X^[(nest)(%s%s)𐶐܂B", "Monster nest (%s%s)"), n_ptr->name, pit_subtype_string(cur_nest_type, TRUE));

	/* Place some monsters */
	for (y = yval - 2; y <= yval + 2; y++)
	{
		for (x = xval - 9; x <= xval + 9; x++)
		{
			MONRACE_IDX r_idx;

			i = randint0(NUM_NEST_MON_TYPE);
			r_idx = nest_mon_info[i].r_idx;

			/* Place that "random" monster (no groups) */
			(void)place_monster_aux(0, y, x, r_idx, 0L);

			nest_mon_info[i].used = TRUE;
		}
	}

	if (cheat_room)
	{
		ang_sort_comp = ang_sort_comp_nest_mon_info;
		ang_sort_swap = ang_sort_swap_nest_mon_info;
		ang_sort(nest_mon_info, NULL, NUM_NEST_MON_TYPE);

		/* Dump the entries (prevent multi-printing) */
		for (i = 0; i < NUM_NEST_MON_TYPE; i++)
		{
			if (!nest_mon_info[i].used) break;
			for (; i < NUM_NEST_MON_TYPE - 1; i++)
			{
				if (nest_mon_info[i].r_idx != nest_mon_info[i + 1].r_idx) break;
				if (!nest_mon_info[i + 1].used) break;
			}
			msg_format_wizard(CHEAT_DUNGEON, "Nest\X^[No.%d:%s", i, r_name + r_info[nest_mon_info[i].r_idx].name);
		}
	}

	return TRUE;
}


/*!
* @brief ^Cv6̕cpit𐶐 / Type 6 -- Monster pits
* @return Ȃ
* @details
* A monster pit is a "big" room, with an "inner" room, containing\n
* a "collection" of monsters of a given type organized in the room.\n
*\n
* The inside room in a monster pit appears as shown below, where the\n
* actual monsters in each location depend on the type of the pit\n
*\n
*   XXXXXXXXXXXXXXXXXXXXX\n
*   X0000000000000000000X\n
*   X0112233455543322110X\n
*   X0112233467643322110X\n
*   X0112233455543322110X\n
*   X0000000000000000000X\n
*   XXXXXXXXXXXXXXXXXXXXX\n
*\n
* Note that the monsters in the pit are now chosen by using "get_mon_num()"\n
* to request 16 "appropriate" monsters, sorting them by level, and using\n
* the "even" entries in this sorted list for the contents of the pit.\n
*\n
* Hack -- all of the "dragons" in a "dragon" pit must be the same "color",\n
* which is handled by requiring a specific "breath" attack for all of the\n
* dragons.  This may include "multi-hued" breath.  Note that "wyrms" may\n
* be present in many of the dragon pits, if they have the proper breath.\n
*\n
* Note the use of the "get_mon_num_prep()" function, and the special\n
* "get_mon_num_hook()" restriction function, to prepare the "monster\n
* allocation table" in such a way as to optimize the selection of\n
* "appropriate" non-unique monsters for the pit.\n
*\n
* Note that the "get_mon_num()" function may (rarely) fail, in which case\n
* the pit will be empty.\n
*\n
* Note that "monster pits" will never contain "unique" monsters.\n
*/
bool build_type6(void)
{
	POSITION y, x, y1, x1, y2, x2, xval, yval;
	int i, j;

	MONRACE_IDX what[16];

	monster_type align;

	cave_type *c_ptr;

	int cur_pit_type = pick_vault_type(pit_types, d_info[dungeon_type].pit);
	vault_aux_type *n_ptr;

	/* No type available */
	if (cur_pit_type < 0) return FALSE;

	n_ptr = &pit_types[cur_pit_type];

	/* Process a preparation function if necessary */
	if (n_ptr->prep_func) (*(n_ptr->prep_func))();

	/* Prepare allocation table */
	get_mon_num_prep(n_ptr->hook_func, NULL);

	align.sub_align = SUB_ALIGN_NEUTRAL;

	/* Pick some monster types */
	for (i = 0; i < 16; i++)
	{
		MONRACE_IDX r_idx = 0;
		int attempts = 100;
		monster_race *r_ptr = NULL;

		while (attempts--)
		{
			/* Get a (hard) monster type */
			r_idx = get_mon_num(dun_level + 11);
			r_ptr = &r_info[r_idx];

			/* Decline incorrect alignment */
			if (monster_has_hostile_align(&align, 0, 0, r_ptr)) continue;

			/* Accept this monster */
			break;
		}

		/* Notice failure */
		if (!r_idx || !attempts) return FALSE;

		/* Note the alignment */
		if (r_ptr->flags3 & RF3_EVIL) align.sub_align |= SUB_ALIGN_EVIL;
		if (r_ptr->flags3 & RF3_GOOD) align.sub_align |= SUB_ALIGN_GOOD;

		what[i] = r_idx;
	}

	/* Find and reserve some space in the dungeon.  Get center of room. */
	if (!find_space(&yval, &xval, 11, 25)) return FALSE;

	/* Large room */
	y1 = yval - 4;
	y2 = yval + 4;
	x1 = xval - 11;
	x2 = xval + 11;

	/* Place the floor area */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		for (x = x1 - 1; x <= x2 + 1; x++)
		{
			c_ptr = &cave[y][x];
			place_floor_grid(c_ptr);
			c_ptr->info |= (CAVE_ROOM);
		}
	}

	/* Place the outer walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		c_ptr = &cave[y][x1 - 1];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y][x2 + 1];
		place_outer_grid(c_ptr);
	}
	for (x = x1 - 1; x <= x2 + 1; x++)
	{
		c_ptr = &cave[y1 - 1][x];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y2 + 1][x];
		place_outer_grid(c_ptr);
	}

	/* Advance to the center room */
	y1 = y1 + 2;
	y2 = y2 - 2;
	x1 = x1 + 2;
	x2 = x2 - 2;

	/* The inner walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		c_ptr = &cave[y][x1 - 1];
		place_inner_grid(c_ptr);
		c_ptr = &cave[y][x2 + 1];
		place_inner_grid(c_ptr);
	}
	for (x = x1 - 1; x <= x2 + 1; x++)
	{
		c_ptr = &cave[y1 - 1][x];
		place_inner_grid(c_ptr);
		c_ptr = &cave[y2 + 1][x];
		place_inner_grid(c_ptr);
	}
	for (y = y1; y <= y2; y++)
	{
		for (x = x1; x <= x2; x++)
		{
			add_cave_info(y, x, CAVE_ICKY);
		}
	}

	/* Place a secret door */
	switch (randint1(4))
	{
	case 1: place_secret_door(y1 - 1, xval, DOOR_DEFAULT); break;
	case 2: place_secret_door(y2 + 1, xval, DOOR_DEFAULT); break;
	case 3: place_secret_door(yval, x1 - 1, DOOR_DEFAULT); break;
	case 4: place_secret_door(yval, x2 + 1, DOOR_DEFAULT); break;
	}

	/* Sort the entries */
	for (i = 0; i < 16 - 1; i++)
	{
		/* Sort the entries */
		for (j = 0; j < 16 - 1; j++)
		{
			int i1 = j;
			int i2 = j + 1;

			int p1 = r_info[what[i1]].level;
			int p2 = r_info[what[i2]].level;

			/* Bubble */
			if (p1 > p2)
			{
				MONRACE_IDX tmp = what[i1];
				what[i1] = what[i2];
				what[i2] = tmp;
			}
		}
	}

	msg_format_wizard(CHEAT_DUNGEON, _("X^[(pit)(%s%s)𐶐܂B", "Monster pit (%s%s)"), n_ptr->name, pit_subtype_string(cur_pit_type, FALSE));

	/* Select the entries */
	for (i = 0; i < 8; i++)
	{
		/* Every other entry */
		what[i] = what[i * 2];
		msg_format_wizard(CHEAT_DUNGEON, _("Nest\X^[INo.%d:%s", "Nest Monster Select No.%d:%s"), i, r_name + r_info[what[i]].name);
	}

	/* Top and bottom rows */
	for (x = xval - 9; x <= xval + 9; x++)
	{
		place_monster_aux(0, yval - 2, x, what[0], PM_NO_KAGE);
		place_monster_aux(0, yval + 2, x, what[0], PM_NO_KAGE);
	}

	/* Middle columns */
	for (y = yval - 1; y <= yval + 1; y++)
	{
		place_monster_aux(0, y, xval - 9, what[0], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 9, what[0], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 8, what[1], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 8, what[1], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 7, what[1], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 7, what[1], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 6, what[2], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 6, what[2], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 5, what[2], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 5, what[2], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 4, what[3], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 4, what[3], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 3, what[3], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 3, what[3], PM_NO_KAGE);

		place_monster_aux(0, y, xval - 2, what[4], PM_NO_KAGE);
		place_monster_aux(0, y, xval + 2, what[4], PM_NO_KAGE);
	}

	/* Above/Below the center monster */
	for (x = xval - 1; x <= xval + 1; x++)
	{
		place_monster_aux(0, yval + 1, x, what[5], PM_NO_KAGE);
		place_monster_aux(0, yval - 1, x, what[5], PM_NO_KAGE);
	}

	/* Next to the center monster */
	place_monster_aux(0, yval, xval + 1, what[6], PM_NO_KAGE);
	place_monster_aux(0, yval, xval - 1, what[6], PM_NO_KAGE);

	/* Center monster */
	place_monster_aux(0, yval, xval, what[7], PM_NO_KAGE);

	return TRUE;
}



/*
* Helper function for "trapped monster pit"
*/
static bool vault_aux_trapped_pit(MONRACE_IDX r_idx)
{
	monster_race *r_ptr = &r_info[r_idx];

	/* Validate the monster */
	if (!vault_monster_okay(r_idx)) return (FALSE);

	/* No wall passing monster */
	if (r_ptr->flags2 & (RF2_PASS_WALL | RF2_KILL_WALL)) return (FALSE);

	/* Okay */
	return (TRUE);
}


/*!
* @brief ^Cv13̕cgbvpit̐ / Type 13 -- Trapped monster pits
* @return Ȃ
* @details
* A trapped monster pit is a "big" room with a straight corridor in\n
* which wall opening traps are placed, and with two "inner" rooms\n
* containing a "collection" of monsters of a given type organized in\n
* the room.\n
*\n
* The trapped monster pit appears as shown below, where the actual\n
* monsters in each location depend on the type of the pit\n
*\n
*  XXXXXXXXXXXXXXXXXXXXXXXXX\n
*  X                       X\n
*  XXXXXXXXXXXXXXXXXXXXXXX X\n
*  XXXXX001123454321100XXX X\n
*  XXX0012234567654322100X X\n
*  XXXXXXXXXXXXXXXXXXXXXXX X\n
*  X           ^           X\n
*  X XXXXXXXXXXXXXXXXXXXXXXX\n
*  X X0012234567654322100XXX\n
*  X XXX001123454321100XXXXX\n
*  X XXXXXXXXXXXXXXXXXXXXXXX\n
*  X                       X\n
*  XXXXXXXXXXXXXXXXXXXXXXXXX\n
*\n
* Note that the monsters in the pit are now chosen by using "get_mon_num()"\n
* to request 16 "appropriate" monsters, sorting them by level, and using\n
* the "even" entries in this sorted list for the contents of the pit.\n
*\n
* Hack -- all of the "dragons" in a "dragon" pit must be the same "color",\n
* which is handled by requiring a specific "breath" attack for all of the\n
* dragons.  This may include "multi-hued" breath.  Note that "wyrms" may\n
* be present in many of the dragon pits, if they have the proper breath.\n
*\n
* Note the use of the "get_mon_num_prep()" function, and the special\n
* "get_mon_num_hook()" restriction function, to prepare the "monster\n
* allocation table" in such a way as to optimize the selection of\n
* "appropriate" non-unique monsters for the pit.\n
*\n
* Note that the "get_mon_num()" function may (rarely) fail, in which case\n
* the pit will be empty.\n
*\n
* Note that "monster pits" will never contain "unique" monsters.\n
*/
bool build_type13(void)
{
	static int placing[][3] = {
		{ -2, -9, 0 },{ -2, -8, 0 },{ -3, -7, 0 },{ -3, -6, 0 },
		{ +2, -9, 0 },{ +2, -8, 0 },{ +3, -7, 0 },{ +3, -6, 0 },
		{ -2, +9, 0 },{ -2, +8, 0 },{ -3, +7, 0 },{ -3, +6, 0 },
		{ +2, +9, 0 },{ +2, +8, 0 },{ +3, +7, 0 },{ +3, +6, 0 },
		{ -2, -7, 1 },{ -3, -5, 1 },{ -3, -4, 1 },
		{ +2, -7, 1 },{ +3, -5, 1 },{ +3, -4, 1 },
		{ -2, +7, 1 },{ -3, +5, 1 },{ -3, +4, 1 },
		{ +2, +7, 1 },{ +3, +5, 1 },{ +3, +4, 1 },
		{ -2, -6, 2 },{ -2, -5, 2 },{ -3, -3, 2 },
		{ +2, -6, 2 },{ +2, -5, 2 },{ +3, -3, 2 },
		{ -2, +6, 2 },{ -2, +5, 2 },{ -3, +3, 2 },
		{ +2, +6, 2 },{ +2, +5, 2 },{ +3, +3, 2 },
		{ -2, -4, 3 },{ -3, -2, 3 },
		{ +2, -4, 3 },{ +3, -2, 3 },
		{ -2, +4, 3 },{ -3, +2, 3 },
		{ +2, +4, 3 },{ +3, +2, 3 },
		{ -2, -3, 4 },{ -3, -1, 4 },
		{ +2, -3, 4 },{ +3, -1, 4 },
		{ -2, +3, 4 },{ -3, +1, 4 },
		{ +2, +3, 4 },{ +3, +1, 4 },
		{ -2, -2, 5 },{ -3, 0, 5 },{ -2, +2, 5 },
		{ +2, -2, 5 },{ +3, 0, 5 },{ +2, +2, 5 },
		{ -2, -1, 6 },{ -2, +1, 6 },
		{ +2, -1, 6 },{ +2, +1, 6 },
		{ -2, 0, 7 },{ +2, 0, 7 },
		{ 0, 0, -1 }
	};

	POSITION y, x, y1, x1, y2, x2, xval, yval;
	int i, j;

	MONRACE_IDX what[16];

	monster_type align;

	cave_type *c_ptr;

	int cur_pit_type = pick_vault_type(pit_types, d_info[dungeon_type].pit);
	vault_aux_type *n_ptr;

	/* Only in Angband */
	if (dungeon_type != DUNGEON_ANGBAND) return FALSE;

	/* No type available */
	if (cur_pit_type < 0) return FALSE;

	n_ptr = &pit_types[cur_pit_type];

	/* Process a preparation function if necessary */
	if (n_ptr->prep_func) (*(n_ptr->prep_func))();

	/* Prepare allocation table */
	get_mon_num_prep(n_ptr->hook_func, vault_aux_trapped_pit);

	align.sub_align = SUB_ALIGN_NEUTRAL;

	/* Pick some monster types */
	for (i = 0; i < 16; i++)
	{
		MONRACE_IDX r_idx = 0;
		int attempts = 100;
		monster_race *r_ptr = NULL;

		while (attempts--)
		{
			/* Get a (hard) monster type */
			r_idx = get_mon_num(dun_level + 0);
			r_ptr = &r_info[r_idx];

			/* Decline incorrect alignment */
			if (monster_has_hostile_align(&align, 0, 0, r_ptr)) continue;

			/* Accept this monster */
			break;
		}

		/* Notice failure */
		if (!r_idx || !attempts) return FALSE;

		/* Note the alignment */
		if (r_ptr->flags3 & RF3_EVIL) align.sub_align |= SUB_ALIGN_EVIL;
		if (r_ptr->flags3 & RF3_GOOD) align.sub_align |= SUB_ALIGN_GOOD;

		what[i] = r_idx;
	}

	/* Find and reserve some space in the dungeon.  Get center of room. */
	if (!find_space(&yval, &xval, 13, 25)) return FALSE;

	/* Large room */
	y1 = yval - 5;
	y2 = yval + 5;
	x1 = xval - 11;
	x2 = xval + 11;

	/* Fill with inner walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		for (x = x1 - 1; x <= x2 + 1; x++)
		{
			c_ptr = &cave[y][x];
			place_inner_grid(c_ptr);
			c_ptr->info |= (CAVE_ROOM);
		}
	}

	/* Place the floor area 1 */
	for (x = x1 + 3; x <= x2 - 3; x++)
	{
		c_ptr = &cave[yval - 2][x];
		place_floor_grid(c_ptr);
		add_cave_info(yval - 2, x, CAVE_ICKY);

		c_ptr = &cave[yval + 2][x];
		place_floor_grid(c_ptr);
		add_cave_info(yval + 2, x, CAVE_ICKY);
	}

	/* Place the floor area 2 */
	for (x = x1 + 5; x <= x2 - 5; x++)
	{
		c_ptr = &cave[yval - 3][x];
		place_floor_grid(c_ptr);
		add_cave_info(yval - 3, x, CAVE_ICKY);

		c_ptr = &cave[yval + 3][x];
		place_floor_grid(c_ptr);
		add_cave_info(yval + 3, x, CAVE_ICKY);
	}

	/* Corridor */
	for (x = x1; x <= x2; x++)
	{
		c_ptr = &cave[yval][x];
		place_floor_grid(c_ptr);
		c_ptr = &cave[y1][x];
		place_floor_grid(c_ptr);
		c_ptr = &cave[y2][x];
		place_floor_grid(c_ptr);
	}

	/* Place the outer walls */
	for (y = y1 - 1; y <= y2 + 1; y++)
	{
		c_ptr = &cave[y][x1 - 1];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y][x2 + 1];
		place_outer_grid(c_ptr);
	}
	for (x = x1 - 1; x <= x2 + 1; x++)
	{
		c_ptr = &cave[y1 - 1][x];
		place_outer_grid(c_ptr);
		c_ptr = &cave[y2 + 1][x];
		place_outer_grid(c_ptr);
	}

	/* Random corridor */
	if (one_in_(2))
	{
		for (y = y1; y <= yval; y++)
		{
			place_floor_bold(y, x2);
			place_solid_bold(y, x1 - 1);
		}
		for (y = yval; y <= y2 + 1; y++)
		{
			place_floor_bold(y, x1);
			place_solid_bold(y, x2 + 1);
		}
	}
	else
	{
		for (y = yval; y <= y2 + 1; y++)
		{
			place_floor_bold(y, x1);
			place_solid_bold(y, x2 + 1);
		}
		for (y = y1; y <= yval; y++)
		{
			place_floor_bold(y, x2);
			place_solid_bold(y, x1 - 1);
		}
	}

	/* Place the wall open trap */
	cave[yval][xval].mimic = cave[yval][xval].feat;
	cave[yval][xval].feat = feat_trap_open;

	/* Sort the entries */
	for (i = 0; i < 16 - 1; i++)
	{
		/* Sort the entries */
		for (j = 0; j < 16 - 1; j++)
		{
			int i1 = j;
			int i2 = j + 1;

			int p1 = r_info[what[i1]].level;
			int p2 = r_info[what[i2]].level;

			/* Bubble */
			if (p1 > p2)
			{
				MONRACE_IDX tmp = what[i1];
				what[i1] = what[i2];
				what[i2] = tmp;
			}
		}
	}

	msg_format_wizard(CHEAT_DUNGEON, _("%s%s㩃sbg܂B", "Trapped monster pit (%s%s)"),
		n_ptr->name, pit_subtype_string(cur_pit_type, FALSE));

	/* Select the entries */
	for (i = 0; i < 8; i++)
	{
		/* Every other entry */
		what[i] = what[i * 2];

		if (cheat_hear)
		{
			/* Message */
			msg_print(r_name + r_info[what[i]].name);
		}
	}

	for (i = 0; placing[i][2] >= 0; i++)
	{
		y = yval + placing[i][0];
		x = xval + placing[i][1];
		place_monster_aux(0, y, x, what[placing[i][2]], PM_NO_KAGE);
	}

	return TRUE;
}

